"""
Webcomics Collector module.
"""

import re
import urllib2
import os
import sys
import gdbm
import ConfigParser
import logging
import logging.handlers
import cPickle
import md5
import time
from urlparse import urljoin
from string import replace


def validateUrl(url):
	valid = re.match(r"(http|https|ftp)://[^ \">\t\r\n]*", url)
	if not valid:
		return False
	else:
		return True


class Collector:
	"""Collector class."""
	def __init__(self, options, args, version):
		"""Initialize class, set variables and start needed action."""
		self.options = options
		self.version = version
		self.initConfig()
		self.initLogger()
		self.initComics()
		
		if self.options.action == "list":
			self.listComics()
		elif self.options.action == "listall":
			self.listAvailableComics()
		elif self.options.action == "add":
			self.addComic(args)
		elif self.options.action == "remove":
			self.removeComic(args)
		elif self.options.action == "backlog":
			self.fetchArchive(args, backlog=True)
		elif self.options.action == "archive":
			self.fetchArchive(args)
		else:
			self.fetchNewest(args)

	def initConfig(self):
		"""Initialize the configuration. Create a clean config if none is found and load the config if it is there already."""
		if os.name == "posix":
			self.cfgdir = os.path.expanduser("~/.collector/")
		elif os.name == "nt":
			self.cfgdir = os.path.expanduser("~\\collector\\")
		else:
			raise UnsupportedOS()
		
		if not os.path.isdir(self.cfgdir):
			# Make the config directory
			os.mkdir(self.cfgdir)

		if not os.path.isfile(self.cfgdir + "collector.cfg"):
			# Create a config file
			self.config = ConfigParser.ConfigParser()
			self.config.add_section("Main")
			self.config.set("Main", "comics", "")
			self.config.set("Main", "comicsdir", self.cfgdir)
			self.config.write(open(self.cfgdir + "collector.cfg", "w"))

		self.config = ConfigParser.ConfigParser()
		self.config.read(self.cfgdir + "collector.cfg")
		if not self.config.has_section("Main"):
			raise ConfigurationError(self.cfgdir + _("collector.cfg does not have a Main section."))
		if not self.config.has_option("Main", "comics"):
			raise ConfigurationError(_("No 'comics' option set in collector.cfg."))
		if not self.config.has_option("Main", "comicsdir"):
			raise ConfigurationError(_("No 'comicsdir' option set in collector.cfg."))
		if self.config.has_option("Main", "useragent"):
			self.useragent = self.config.get("Main", "useragent")
		else:
			self.useragent = "WebcomicsCollector/" + self.version
		if self.config.has_option("Main", "logfile"):
			self.logfile = os.path.expanduser(self.config.get("Main", "logfile"))
		else:
			self.logfile = self.cfgdir + "log"
		if self.config.has_option("Main", "loglevel"):
			self.loglevel = self.config.get("Main", "loglevel")
			if self.loglevel not in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"):
				raise ConfigurationError(_("Invalid value for loglevel: %s") % str(self.loglevel))
		else:
			self.loglevel = "WARNING"
	
	def initLogger(self):
		"""Set up the logger."""
		self.logtext = ""
		self.logger = logging.getLogger("collector")
		if self.logfile == "SYSLOG":
			hdlr = logging.handlers.SysLogHandler()
			formatter = logging.Formatter("Collector: %(message)s")
			hdlr.setFormatter(formatter)
			self.logger.addHandler(hdlr)
		else:
			hdlr = logging.FileHandler(self.logfile)
			formatter = logging.Formatter("%(asctime)s %(message)s")
			hdlr.setFormatter(formatter)
			self.logger.addHandler(hdlr)
		if self.loglevel == "DEBUG":
			self.logger.setLevel(logging.DEBUG)
		elif self.loglevel == "INFO":
			self.logger.setLevel(logging.INFO)
		elif self.loglevel == "WARNING":
			self.logger.setLevel(logging.WARNING)
		elif self.loglevel == "ERROR":
			self.logger.setLevel(logging.ERROR)
		elif self.loglevel == "CRITICAL":
			self.logger.setLevel(logging.CRITICAL)
		if sys.stdout.isatty():
			self.ttylogger = logging.getLogger("tty")
			hdlr = logging.StreamHandler()
			formatter = logging.Formatter("%(asctime)s %(message)s")
			hdlr.setFormatter(formatter)
			self.ttylogger.addHandler(hdlr)
			if self.options.verbose == "quiet":
				self.ttylogger.setLevel(logging.CRITICAL)
			elif self.options.verbose == "verbose":
				self.ttylogger.setLevel(logging.INFO)
			elif self.options.verbose == "debug":
				self.ttylogger.setLevel(logging.DEBUG)
			else:
				self.ttylogger.setLevel(logging.WARNING)
		else:
			self.ttylogger = False

	def initComics(self):
		"""Initialize the comics definitions."""
		if os.name == "posix":
			defdirs = ("/etc/collector/", "/usr/local/etc/collector/", self.cfgdir)
		elif os.name == "nt":
			defdirs = (self.cfgdir,)
		else:
			raise UnsupportedOS()
		
		# Load comics classes
		# Check if classes have been cached and get mtime
		classes_cache_filename = self.cfgdir + "comics.classes.cache"
		if os.path.isfile(classes_cache_filename):
			classes_cache_mtime = os.path.getmtime(classes_cache_filename)
		else:
			classes_cache_mtime = 0
		loadcache = True
		for dir in defdirs:
			filename = dir + "comics.classes"
			if os.path.isfile(filename):
				if classes_cache_mtime < os.path.getmtime(filename):
					loadcache = False
		if loadcache:
			self.comicsclasses = cPickle.load(open(classes_cache_filename, "r"))
		else:
			self.comicsclasses = ConfigParser.ConfigParser()
			for dir in defdirs:
				self.comicsclasses.read(dir + "comics.classes")
			cPickle.dump(self.comicsclasses, open(classes_cache_filename, "w"), cPickle.HIGHEST_PROTOCOL)

		# Load comics definitions
		def_cache_filename = self.cfgdir + "comics.def.cache"
		if os.path.isfile(def_cache_filename):
			def_cache_mtime = os.path.getmtime(def_cache_filename)
		else:
			def_cache_mtime = 0
		loadcache = True
		for dir in defdirs:
			filename = dir + "comics.def"
			if os.path.isfile(filename):
				if def_cache_mtime < os.path.getmtime(filename):
					loadcache = False
		if loadcache:
			self.comicsdef = cPickle.load(open(def_cache_filename, "r"))
		else:
			self.comicsdef = ConfigParser.ConfigParser()
			for dir in defdirs:
				self.comicsdef.read(dir + "comics.def")
			cPickle.dump(self.comicsdef, open(def_cache_filename, "w"), cPickle.HIGHEST_PROTOCOL)

		self.comicsdir = os.path.expanduser(self.config.get("Main", "comicsdir"))
		if self.comicsdir[-1] != os.sep:
			self.comicsdir += os.sep

		self.comics = []
		for comic in self.config.get("Main", "comics").split(","):
			comic = comic.strip()
			if comic != "":
				if self.comicsdef.has_section(comic):
					self.log("DEBUG", _("Initializing comic '%s'") % comic)
					options = self.getComicConfig(comic)
					self.comics.append(Comic(comic, options, self.log, self.comicsdir, self.useragent, self.options.redownload))
				else:
					self.log("WARNING", _("%s was in comicslist, but no definition was found for it.") % comic)
					self.removeComic([comic])

	def getComicConfig(self, comic):
		"""Get config for comic from comics.def + comics.classes."""
		options = {}
		if self.comicsdef.has_option(comic, "class"):
			try:
				for opt in self.comicsclasses.items(self.comicsdef.get(comic, "class")):
					options[opt[0]] = opt[1]
			except ConfigParser.NoSectionError, error:
				raise ConfigurationError(_("The definition for '%s' references non-existing class '%s'") % (comic, error.section))
		for opt in self.comicsdef.items(comic):
			options[opt[0]] = opt[1]
		for opt in options:
			matches = None
			while matches != []:
				matches = re.findall("\{\{(.+?)\}\}", options[opt])
				for match in matches:
					if match == "ext":
						options[opt] = re.sub("\{\{" + match + "\}\}", "(?:jpg|jpeg|jpe|gif|png|bmp)", options[opt])
					elif match == "signs":
						options[opt] = re.sub("\{\{" + match + "\}\}", "[a-z0-9_\-\.%]+?", options[opt])
					elif match == "filename":
						options[opt] = re.sub("\{\{" + match + "\}\}", "[a-z0-9_\-\.%]+?\.(?:jpg|jpeg|jpe|gif|png|bmp)", options[opt])
					elif opt == "archiveurl" and match[:5] == "DATE:":
						date = time.strftime(match[5:])
						url = re.sub("\{\{" + match + "\}\}", date, options[opt])
						try:
							urllib2.urlopen(url)
						except urllib2.HTTPError:
							date = time.strftime(match[5:], time.localtime(time.time() - 86400))
							url = re.sub("\{\{" + match + "\}\}", date, options[opt])
						options[opt] = url
					else:
						try:
							options[opt] = re.sub("\{\{" + match + "\}\}", options[match], options[opt])
						except KeyError:
							raise ConfigurationError(_("Invalid substitution '{{%s}}' in %s for %s, or it's associated class.") % (match, opt, comic))
		return options

	def log(self, level, text, endline = True):
		"""Log text to logger instance."""
		self.logtext += text
		if endline:
			if level == "DEBUG":
				self.logger.debug(self.logtext)
				if self.ttylogger:
					self.ttylogger.debug(self.logtext)
			if level == "INFO":
				self.logger.info(self.logtext)
				if self.ttylogger:
					self.ttylogger.info(self.logtext)
			if level == "WARNING":
				self.logger.warning(self.logtext)
				if self.ttylogger:
					self.ttylogger.warning(self.logtext)
			if level == "ERROR":
				self.logger.error(self.logtext)
				if self.ttylogger:
					self.ttylogger.error(self.logtext)
			if level == "CRITICAL":
				self.logger.critical(self.logtext)
				if self.ttylogger:
					self.ttylogger.critical(self.logtext)
			self.logtext = ""

	def saveConfig(self):
		"""Update config object and save it to collector.cfg."""
		self.log("INFO", _("Saving config."))
		if self.comics == []:
			comicslist = ""
		else:
			comicslist = self.comics[0].name
			for comic in self.comics[1:]:
				comicslist += ", " + comic.name
		self.config.set("Main", "comics", comicslist)
		self.config.write(open(self.cfgdir + "collector.cfg", "w"))
	
	def listAvailableComics(self):
		"""List all known comics"""
		self.log("DEBUG", _("Listing all known comics."))
		comics = self.comicsdef.sections()
		comics.sort()
		output = ""
		for comic in comics:
			options = self.getComicConfig(comic)
			output += "\"" + comic + "\" (" + options["homepage"] + ")\n"
		print _("These comics are known and available for adding to your comicslist:")
		print output

	def listComics(self):
		"""List users comicslist"""
		self.log("DEBUG", _("Listing comics on comicslist."))
		if self.comics == []:
			raise NoComicsError()
		else:
			print _("These comics are on your comicslist:")
			for comic in self.comics:
				print "\"" + comic.name + "\" (" + comic.homepage + ")"

	def configuredComics(self):
		"""Return users comicslist as a list."""
		names = []
		for comic in self.comics:
			names.append(comic.name)
		return names

	def removeComic(self, comiclist):
		"""Remove comic from users comicslist."""
		if comiclist == [] or comiclist is None:
			# No comics specified, therefore all comics are implied.
			comiclist = self.configuredComics()
		for comicname in comiclist:
			if comicname in self.configuredComics():
				self.log("INFO", _("Removing comic '%s' from comicslist.") % comicname)
				for comic in self.comics:
					if comic.name == comicname:
						self.comics.remove(comic)
			else:
				self.log("WARNING", _("Comic '%s' not found in comicslist, can't remove it.") % comicname)
		self.saveConfig()

	def addComic(self, comiclist):
		"""Add comic(s) to users comicslist."""
		if comiclist == [] or comiclist is None:
			# No comics specified, therefore all comics are implied.
			comiclist = self.comicsdef.sections()

		comiclist.sort()
		for comic in comiclist:
			if comic in self.comicsdef.sections():
				if comic not in self.configuredComics():
					self.log("INFO", _("Adding comic '%s' to your comicslist.") % comic)
					options = self.getComicConfig(comic)
					self.comics.append(Comic(comic, options, self.log, self.comicsdir, self.useragent, self.options.redownload))
					self.saveConfig()
				else:
					self.log("WARNING", _("Can not add '%s', it is already in your comicslist.") % comic)
			else:
				self.log("WARNING", _("'%s' not found in comics.def, not adding.") % comic)

	def fetchNewest(self, comiclist):
		"""Fetch newest strip for specified comic(s) or for all if no comic(s) specified."""
		if self.comics == []:
			raise NoComicsError()
		if comiclist == [] or comiclist is None:
			self.log("INFO", _("No comics specified, fetching newest for all comics in comicslist."))
			comiclist = self.configuredComics()
		errors = []
		for comic in comiclist:
			if comic in self.configuredComics():
				for comicobj in self.comics:
					if comic == comicobj.name:
						self.log("INFO", _("[%s] Fetching newest strip(s).") % comic)
						try:
							comicobj.fetchNewest()
						except DownloadError, error:
							errors.append((comic, error))
						except DownloadErrors, error:
							errors.append((comic, error))
						except MissingOption, error:
							errors.append((comic, error))
						except ConfigurationError, error:
							errors.append((comic, error))
						except RegExpError, error:
							errors.append((comic, error))
						except RegExpNoResultError, error:
							errors.append((comic, error))
						except InvalidUrlError, error:
							errors.append((comic, error))
			else:
				self.log("WARNING", _("'%s' isn't in your comicslist, not fetching newest strips for it.") % comic)
		if errors != []:
			raise ComicErrors(errors)

	def fetchArchive(self, comiclist, backlog=False):
		"""Fetch archive or backlog for specified comic(s) or for all if no comic(s) specified."""
		if self.comics == []:
			raise NoComicsError()
		if comiclist == [] or comiclist is None:
			comiclist = self.configuredComics()
		errors = []
		for comic in comiclist:
			if comic in self.configuredComics():
				for comicobj in self.comics:
					if comic == comicobj.name:
						if backlog:
							self.log("INFO", _("[%s] Fetching backlog.") % comic)
						else:
							self.log("INFO", _("[%s] Fetching archive.") % comic)
						try:
							comicobj.fetchArchive(backlog=backlog)
						except DownloadError, error:
							errors.append((comic, error))
						except DownloadErrors, error:
							errors.append((comic, error))
						except MissingOption, error:
							errors.append((comic, error))
						except ConfigurationError, error:
							errors.append((comic, error))
						except RegExpError, error:
							errors.append((comic, error))
						except RegExpNoResultError, error:
							errors.append((comic, error))
						except InvalidUrlError, error:
							errors.append((comic, error))
			else:
				if backlog:
					self.log("WARNING", _("'%s' isn't in your comicslist, not fetching backlog for it.") % comic)
				else:
					self.log("WARNING", _("'%s' isn't in your comicslist, not fetching archive for it.") % comic)
		if errors != []:
			raise ComicErrors(errors)


class Comic:
	"""Comic class."""
	def __init__(self, name, options, log, comicsdir, useragent, redownload):
		"""Set all options as attributes of the class."""
		self.name = name
		self.log = log
		self.dir = comicsdir + self.name + os.sep
		self.useragent = useragent
		self.redownload = redownload
		if not os.path.isdir(self.dir):
			self.log("INFO", _("[%s] %s does not exist, now creating.") % (self.name, self.dir))
			os.mkdir(self.dir)
		for opt in options:
			setattr(self, opt, options[opt])
		self.initStripsDat()

	def initStripsDat(self):
		"""Reads and parses strips.dat."""
		self.log("DEBUG", _("[%s] Reading strips.dat") % self.name)
		self.strips = []
		if os.path.isfile(self.dir + "strips.dat"):
			for line in open(self.dir + "strips.dat", "r"):
				fields = line.split("|")
				strip = [fields[0]]
				for subfield in fields[1:]:
					filename, md5sum = subfield.split("/")
					strip.append((filename.strip(), md5sum.strip()))
				self.strips.append(strip)
		self.stripfiles = []
		for strip in self.strips:
			files = []
			for filename, md5sum in strip[1:]:
				files.append(filename)
			self.stripfiles.append(files)

	def checkDef(self):
		"""Check all options, raise exceptions if required options are missing and validate what can be validated."""
		self.log("DEBUG", _("[%s] Checking def options.") % self.name)
		# Check type option
		try:
			self.type
		except AttributeError:
			self.log("DEBUG", _("[%s] Option 'type' is missing.") % self.name)
			raise MissingOption("type")
		if self.type != "search" and self.type != "lateststatic":
			self.log("DEBUG", _("[%s] Type '%s' is invalid.") % (self.name, self.type))
			raise ConfigurationError(_("Invalid type '%s' specified for comic '%s'") % (self.type, self.name))

		# Check homepage option
		try:
			if not validateUrl(self.homepage):
				raise InvalidUrlError("homepage", self.homepage)
		except AttributeError:
			self.log("DEBUG", _("[%s] Option 'homepage' is missing.") % self.name)
			raise MissingOption("homepage")

		# Check archiveurl option
		try:
			self.archiveurl = str(self.archiveurl)
		except:
			self.archiveurl = self.homepage
		if not validateUrl(self.archiveurl):
			raise InvalidUrlError("archiveurl", self.archiveurl)

		# Check authors option
		try:
			self.authors = str(self.authors)
		except AttributeError:
			raise MissingOption("authors")

		# Check imgexp option
		try:
			if self.type == "search":
				self.imgexp = re.compile(self.imgexp, re.I)
		except re.error, err:
			raise RegExpError("imgexp", self.imgexp, err)
		except AttributeError:
			raise MissingOption("imgexp")

		# Check imgurl option
		try:
			if self.type == "lateststatic":
				if not validateUrl(self.imgurl):
					raise InvalidUrlError("imgurl", self.imgurl)
		except AttributeError:
			raise MissingOption("imgurl")
		
		# Check altimgexp option
		try:
			if self.type == "search" and self.altimgexp:
				self.altimgexp = re.compile(self.altimgexp, re.I)
		except re.error, err:
			raise RegExpError("altimgexp", self.altimgexp, err)
		except AttributeError:
			pass

		# Check previousexp option
		try:
			if self.type == "search":
				self.previousexp = re.compile(self.previousexp, re.I)
		except re.error, err:
			raise RegExpError("previousexp", self.previousexp, err)
		except AttributeError:
			raise MissingOption("previousexp")
			
		# Check altpreviousexp option
		try:
			if self.type == "search" and self.altpreviousexp:
				self.altpreviousexp = re.compile(self.altpreviousexp, re.I)
		except re.error, err:
			raise RegExpError("altpreviousexp", self.altpreviousexp, err)
		except AttributeError:
			pass
	
	def __str__(self):
		"""Return string representation of comic (the name)."""
		return self.name

	def writeStripsDat(self):
		"""Write the strips.dat file."""
		self.log("INFO", _("[%s] Writing strips.dat") % self.name)
		data = ""
		for strip in self.strips:
			data += strip[0]
			for filename, md5sum in strip[1:]:
				data += "|" + filename + "/" + md5sum + "\n"
		open(self.dir + "strips.dat", "w").write(data)

	def download(self, url, referer=None):
		"""Downloads the specified URL with the specified referer and returns the content."""
		request = urllib2.Request(url)
		if referer is not None:
			request.add_header("Referer", referer)
		request.add_header("User-Agent", self.useragent)
		try:
			file = urllib2.urlopen(request).read()
		except urllib2.HTTPError, error:
			raise DownloadError(url, "[%s] %s" % (str(error.code), str(error.msg)))
		except urllib2.URLError, error:
			raise DownloadError(url, _("[URL Error] %s") % str(error.reason.args[1]))
		except ValueError, error:
			raise DownloadError(url, _("[Invalid URL] %s") % str(error.args[0]))
		return file

	def downloadImage(self, files, referer):
		"""
		Download image(s) and places it/them in comicdir.
		Takes a list of two-part tuples containing url and name.
		Does not return anything but raises exceptions of anything goes wrong.
		"""
		for f in files:
			url, name = f
			if not os.path.isfile(self.dir + name) or self.redownload:
				# Download file if it does not exist or if redownload option is set
				self.log("INFO", _("[%s] Downloading file: %s") % (self.name, name))
				img = self.download(url, referer)
				open(self.dir + name, "w").write(img)
			else:
				self.log("INFO", _("[%s] Not downloading file (already downloaded): %s") % (self.name, name))
	
	def downloadHtml(self, url, referer=None):
		"""
		Downloads the specified URL with the specified referer and strips tabs 
		and newlines and excessive whitespace before returning the content.
		"""
		html = self.download(url, referer)
		html = re.sub("\n|\r|\t", "", html)
		html = re.sub(">\s+<", "><", html)
		return html

	def applyArchiveExp(self):
		"""If archiveexp exists, archiveurl will be searched with it and the result joined with archiveurl."""
		try:
			self.archiveexp
			self.log("INFO", _("[%s] Finding archiveurl: ") % self.name, endline=False)
			html = self.downloadHtml(self.archiveurl)
			matches = re.search(self.archiveexp, html)
			if matches is not None:
				self.archiveurl = urljoin(self.archiveurl, matches.groups()[0])
				self.log("INFO", self.archiveurl)
				del self.archiveexp
			else:
				raise RegExpNoResultError(self.archiveurl, self.archiveexp)
		except AttributeError:
			pass

	def applyImgExp(self, html, url):
		"""Search specified html using the imgexp regexp and return the results."""
		strips = re.findall(self.imgexp, html)
		if strips == []:
			# No strips found, try altimgexp instead
			try:
				strips = re.findall(self.altimgexp, html)
			except AttributeError:
				pass
		# Create full URLs for all stripfiles
		for n in range(len(strips)):
			try:
				strips[n] = self.imgurl + str(strips[n])
			except AttributeError:
				strips[n] = urljoin(url, strips[n])
		# Find stripnames
		try:
			names = re.findall(self.imgname, html)
		except AttributeError:
			names = []
			for strip in strips:
				try:
					self.pathlevels
					head, tail = os.path.split(strip)
					name = tail
					for n in range(int(self.pathlevels) - 1):
						head, tail = os.path.split(head)
						name = tail + "_" + name
				except AttributeError:
					name = os.path.basename(strip)
				names.append(name)
		returnlist = []
		for n in range(len(strips)):
			name = replace(names[n], "/", "_")
			returnlist.append((strips[n], name))
		return returnlist

	def applyPreviousExp(self, html, url):
		"""Search specified html using the previousexp regexp and return the results."""
		matches = re.findall(self.previousexp, html)
		if matches == []:
			# No matches found, try altpreviousexp instead
			try:
				matches = re.findall(self.altpreviousexp, html)
			except AttributeError:
				pass
		if matches != []:
			prevstrip = urljoin(url, str(matches[0]))
		else:
			prevstrip = ""
		return prevstrip

	def fetchNewest(self):
		"""Determine comic type and call proper method to fetch the newest strip."""
		self.checkDef()
		# Find archiveurl if needed
		self.applyArchiveExp()
		if self.type == "search":
			self.fetchNewestTypeSearch()
		elif self.type == "lateststatic":
			self.fetchNewestTypeLateststatic()

	def fetchNewestTypeLateststatic(self):
		"""Fetch newest strip for comic of type "lasteststatic"."""
		lateststripname = int(self.strips[-1][1][0])
		newstripname = str(lateststripname + 1)
		img = self.download(self.imgurl, self.archiveurl)
		md5sum = md5.new()
		md5sum.update(img)
		md5sum = md5sum.hexdigest()
		md5sum_found = False
		for strip in self.strips:
			for filename in strip[1:]:
				if md5sum == filename[1]:
					md5sum_found = True
		if not md5sum_found:
			self.log("DEBUG", _("[%s] md5sum not found in strips.dat, saving file.") % self.name)
			open(self.dir + str(newstripname), "w").write(img)
			self.strips.append([self.archiveurl, (newstripname, md5sum)])
			self.writeStripsDat()
		elif self.redownload:
			self.log("WARNING", _("[%s] redownload option set, but this is a lateststatic webcomic, ignoring.") % self.name)
		else:
			self.log("DEBUG", _("[%s] md5sum already in strips.dat, not saving file.") % self.name)

	def fetchNewestTypeSearch(self):
		"""Fetch newest strip of type for comic of type "search"."""
		self.log("INFO", _("[%s] Finding newest strip(s): ") % self.name, endline=False)
		if not validateUrl(self.archiveurl):
			raise InvalidUrlError("archiveurl", self.archiveurl)

		# Grab page
		html = self.downloadHtml(self.archiveurl)

		# Find neweststrip(s)
		neweststrips = self.applyImgExp(html, self.archiveurl)

		# Log the names found.
		if neweststrips == []:
			self.log("WARNING", _("No strips found"))
		else:
			self.log("INFO", neweststrips[0][1], endline=False)
			for strip in neweststrips[1:]:
				self.log("INFO", ", " + strip[1], endline=False)
			self.log("INFO", "")

			newfilenames = []
			for url, newfilename in neweststrips:
				newfilenames.append(newfilename)
			alreadydownloaded = False
			for strip in self.strips:
				filenames = []
				for filename, md5sum in strip[1:]:
					filenames.append(filename)
				if newfilenames == filenames:
					alreadydownloaded = True
					break

			# Check whether files have been downloaded
			if not alreadydownloaded or self.redownload:
				self.downloadImage(neweststrips, referer = self.archiveurl)

				# Update URL to previous strip in strips.dat
				if self.strips != []:
					prevstripurl = self.applyPreviousExp(html, self.archiveurl)
					html = self.downloadHtml(prevstripurl, self.archiveurl)
					prevstrips = self.applyImgExp(html, prevstripurl)
					prevstripfilenames = []
					for url, filename in prevstrips:
						prevstripfilenames.append(filename)
					if prevstripfilenames == self.stripfiles[-1]:
						self.strips[-1][0] = prevstripurl

				strip = [self.archiveurl]
				for url, filename in neweststrips:
					md5sum = md5.new()
					md5sum.update(open(self.dir + filename, "r").read())
					md5sum = md5sum.hexdigest()
					strip.append((filename, md5sum))
				self.strips.append(strip)
				self.writeStripsDat()
			else:
				self.log("INFO", _("[%s] Newest strip already downloaded") % self.name)

	def fetchArchive(self, backlog=False):
		"""Download the archived strips. If backlog is True then stop when reaching first already downloaded strip."""
		if self.type == "lateststatic":
			if backlog:
				self.log("WARNING", _("[%s] Can't fetch backlog for comics with type 'lateststatic', fetching latest instead.") % self.name)
			else:
				self.log("WARNING", _("[%s] Can't fetch archive for comics with type 'lateststatic', fetching latest instead.") % self.name)
			self.fetchNewest()
		elif self.type == "search":
			self.checkDef()
			self.applyArchiveExp()
			visitedurls = list()
			strips = list()
			currenturl = self.archiveurl
			prevurl = None
			try:
				while 1:
					self.log("INFO", _("[%s] Finding strip(s): ") % self.name, endline=False)
					html = self.downloadHtml(currenturl, prevurl)
					imgs = self.applyImgExp(html, currenturl)
					if imgs == []:
						self.log("WARNING", _("No strips found."))
					else:
						self.log("INFO", imgs[0][1], endline=False)
						for url, filename in imgs[1:]:
							self.log("INFO", ", " + filename, endline=False)
						self.log("INFO", "")
						for image in imgs:
							if backlog and os.path.isfile(self.dir + image[1]):
								# Backlog is now cleared, download stops here
								self.log("INFO", _("[%s] %s already exists, stopping download.") % (self.name, image[1]))
								raise BacklogCleared()
						self.downloadImage(imgs, referer = currenturl)
						strip = [currenturl]
						for url, filename in imgs:
							md5sum = md5.new()
							md5sum.update(open(self.dir + filename, "r").read())
							md5sum = md5sum.hexdigest()
							strip.append((filename, md5sum))
						strips.append(strip)
					newurl = self.applyPreviousExp(html, currenturl)
					visitedurls.append(currenturl)
					if newurl == "":
						break
					elif newurl in visitedurls:
						self.log("WARNING", _("[%s] Loop detected, stopping download.") % self.name)
						break
					else:
						prevurl = currenturl
						currenturl = newurl
			except BacklogCleared:
				pass
			except KeyboardInterrupt:
				print "Keyboard Interrupt detected, stopping archive download."
			strips.reverse()
			if backlog:
				self.strips += strips
			else:
				self.strips = strips
			self.writeStripsDat()


class NoComicsError(Exception):
	"""Exception raised if no comics are present on the users comicslist."""
	def __str__(self):
		return _("No comics in your comicslist, please add some!")


class BacklogCleared(Exception):
	"""Exception raised during backlog download when images that have already been downloaded are reached."""
	pass


class RegExpError(Exception):
	"""Exception raised if a regular expression contains any errors."""
	def __init__(self, option, exp, experror):
		self.option = option
		self.exp = exp
		self.experror = str(experror)
	def __str__(self):
		return _("The regexp in %s, %s, contains this error: %s") % (self.option, self.exp, self.experror)


class RegExpNoResultError(Exception):
	"""Exception raised if a regular expression returns no results."""
	def __init__(self, url, exp):
		self.url = url
		self.exp = exp
	def __str__(self):
		return _("The regexp '%s' returned no results when applied to %s") % (self.exp, self.url)


class DownloadError(Exception):
	"""Exception raised if a file could not be downloaded."""
	def __init__(self, url, errormsg):
		self.url = url
		self.errormsg = errormsg
	def __str__(self):
		return self.url + ": " + self.errormsg


class ConfigurationError(Exception):
	"""Exception raised if a configuration error has been encountered."""
	def __init__(self, errormsg):
		self.errormsg = errormsg
	def __str__(self):
		return self.errormsg


class MissingOption(Exception):
	"""Exception raised if a required option in a def is missing."""
	def __init__(self, option):
		self.option = option
	def __str__(self):
		return _("The required option %s is missing from comic definition.") % self.option


class InvalidUrlError(Exception):
	"""Exception raised if a URL specified in a def is invalid."""
	def __init__(self, option, url):
		self.option = option
		self.url = url
	def __str__(self):
		return _("Option %s contains invalid URL: %s") % (self.option, self.url)


class DownloadErrors(Exception):
	"""Exception raised if there was any download errors."""
	def __init__(self, errors):
		self.errors = errors
	def __str__(self):
		errormsg = ""
		for error in self.errors:
			errormsg += error + "\n"
		return errormsg


class ComicErrors(Exception):
	"""Exception raised at the end of a run if any errors occurred."""
	def __init__(self, errors):
		self.errors = errors
	def __str__(self):
		errormsg = ""
		for error in self.errors:
			errormsg += error[0] + ":\n"
			errormsg += str(error[1]) + "\n\n"
		return errormsg

class UnsupportedOS(Exception):
	"""Exception raised when trying to run on an unsupported OS."""
	def __str__(self):
		return _("Your current OS isn't supported by Collector. If you have any programming skills, please consider getting Collector to run on it and submit a patch on http://bugs.skumleren.net.")

